"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt

import backoff
import deprecation
import logging
from time import sleep
from typing import Tuple, Union, List, Dict

from gs_quant.api.api_session import ApiWithCustomSession
from gs_quant.common import PositionType
from gs_quant.common import RiskRequest, Currency
from gs_quant.errors import MqInternalServerError, MqTimeoutError, MqRateLimitedError
from gs_quant.instrument import Instrument
from gs_quant.session import GsSession
from gs_quant.target.portfolios import Portfolio, Position, PositionSet, PortfolioTree
from gs_quant.target.reports import Report
from gs_quant.target.risk_models import RiskModelTerm as Term
from gs_quant.workflow import WorkflowPosition, WorkflowPositionsResponse, SaveQuoteRequest

_logger = logging.getLogger(__name__)


class GsPortfolioApi(ApiWithCustomSession):
    """GS Asset API client implementation.

    To pass additional query parameters, use kwargs."""

    @classmethod
    def get_portfolios(cls,
                       portfolio_ids: List[str] = None,
                       portfolio_names: List[str] = None,
                       limit: int = 100,
                       **kwargs) -> Tuple[Portfolio, ...]:
        url = '/portfolios?'
        if portfolio_ids:
            url += f'&id={"&id=".join(portfolio_ids)}'
        if portfolio_names:
            url += f'&name={"&name=".join(portfolio_names)}'
        for k, v in kwargs.items():
            if isinstance(v, list):
                for i in v:
                    # in case v is not a list of strings
                    url += f'&{k}={i}'
            else:
                url += f'&{k}={v}'
        return GsSession.current._get(f'{url}&limit={limit}', cls=Portfolio)['results']

    @classmethod
    def get_portfolio(cls, portfolio_id: str) -> Portfolio:
        return GsSession.current._get('/portfolios/{id}'.format(id=portfolio_id), cls=Portfolio)

    @classmethod
    def get_portfolio_by_name(cls, name: str) -> Portfolio:
        ret = cls.get_session()._get('/portfolios?name={}'.format(name))
        num_found = ret.get('totalResults', 0)

        if num_found == 0:
            raise ValueError('Portfolio {} not found'.format(name))
        elif num_found > 1:
            raise ValueError('More than one portfolio named {} found'.format(name))
        else:
            return Portfolio.from_dict(ret['results'][0])

    @classmethod
    def create_portfolio(cls, portfolio: Portfolio) -> Portfolio:
        return GsSession.current._post('/portfolios', portfolio, cls=Portfolio)

    @classmethod
    def update_portfolio(cls, portfolio: Portfolio):
        return GsSession.current._put('/portfolios/{id}'.format(id=portfolio.id), portfolio, cls=Portfolio)

    @classmethod
    def delete_portfolio(cls, portfolio_id: str) -> dict:
        return GsSession.current._delete('/portfolios/{id}'.format(id=portfolio_id))

    # manage portfolio positions
    @classmethod
    def get_portfolio_analyze(cls, portfolio_id: str) -> dict:
        url = '/portfolios/{id}/analyze'.format(id=portfolio_id)
        res = GsSession.current._get(url)
        return res

    @classmethod
    @backoff.on_exception(lambda: backoff.expo(base=2, factor=2),
                          (MqTimeoutError, MqInternalServerError),
                          max_tries=5)
    @backoff.on_exception(lambda: backoff.constant(90),
                          MqRateLimitedError,
                          max_tries=5)
    def get_positions(cls, portfolio_id: str, start_date: dt.date = None, end_date: dt.date = None,
                      position_type: str = 'close') -> Tuple[PositionSet, ...]:
        url = '/portfolios/{id}/positions?type={positionType}'.format(id=portfolio_id, positionType=position_type)
        if start_date is not None:
            url += '&startDate={sd}'.format(sd=start_date.isoformat())
        if end_date is not None:
            url += '&endDate={sd}'.format(sd=end_date.isoformat())

        res = GsSession.current._get(url)
        return tuple(PositionSet.from_dict(v) for v in res.get('positionSets', ()))

    @classmethod
    def get_positions_for_date(cls, portfolio_id: str, position_date: dt.date,
                               position_type: str = 'close') -> PositionSet:
        url = '/portfolios/{id}/positions/{date}?type={ptype}'.format(
            id=portfolio_id, date=position_date.isoformat(), ptype=position_type)
        position_sets = GsSession.current._get(url, cls=PositionSet)['results']
        return position_sets[0] if len(position_sets) > 0 else None

    @classmethod
    def get_position_set_by_position_type(cls, positions_type: str, positions_id: str,
                                          activity_type: str = 'position') -> Tuple[PositionSet, ...]:
        root = 'deals' if positions_type == 'ETI' else 'books/' + positions_type
        if activity_type != 'position':
            url = '/risk-internal/{}/{}/positions?activityType={}'.format(root, positions_id, activity_type)
        else:
            url = '/risk-internal/{}/{}/positions'.format(root, positions_id)

        results = cls.get_session()._get(url, timeout=181)
        return tuple(cls._unpack_position_set(res) for res in results['positionSets'])

    @classmethod
    def _unpack_position_set(cls, kvs: dict) -> PositionSet:
        position_set = PositionSet.from_dict(kvs)
        # NY - this is rather unfortunate: we end up with instruments all with the name of the ETI
        for position in position_set.positions:
            position.instrument.name = None

        return position_set

    @classmethod
    def get_instruments_by_position_type(cls, positions_type: str,
                                         positions_id: str,
                                         activity_type: str) -> Tuple[Instrument, ...]:
        position_sets = cls.get_position_set_by_position_type(positions_type=positions_type,
                                                              positions_id=positions_id,
                                                              activity_type=activity_type)

        instruments = []
        for position_set in position_sets:
            for positions in position_set.positions:
                instrument = positions.instrument
                instrument.metadata = {
                    'trade_date': position_set.position_date, 'tags': positions.tags,
                    'external_ids': {id_['idType']: id_['idValue'] for id_ in positions.external_ids},
                    'party_from': positions.party_from, 'party_to': positions.party_to}
                instruments.append(instrument)

        return tuple(instruments)

    @classmethod
    def get_latest_positions(cls, portfolio_id: str, position_type: str = 'close') -> Union[PositionSet, dict]:
        url = '/portfolios/{id}/positions/last?type={ptype}'.format(id=portfolio_id, ptype=position_type)
        results = GsSession.current._get(url)['results']

        # Annoyingly, different types are returned depending on position_type

        if isinstance(results, dict) and 'positions' in results:
            results['positions'] = tuple(Position.from_dict(p) for p in results['positions'])

        return PositionSet.from_dict(results)

    @classmethod
    def get_instruments_by_workflow_id(cls, workflow_id: str,
                                       prefer_instruments: bool = False) -> Tuple[Instrument, ...]:
        root = 'quote'
        url = '/risk{}/{}/{}'.format('-internal' if not prefer_instruments else '', root, workflow_id)
        results = cls.get_session()._get(url, timeout=181)

        instruments = []
        for position in results.get('workflowPositions').get(workflow_id):
            for insts in position['positions']:
                instrument_values = insts['instrument']
                instrument = Instrument.from_dict(instrument_values)
                name = instrument_values.get('name')
                if name:
                    instrument.name = name

                instruments.append(instrument)

        return tuple(instruments)

    @classmethod
    def get_position_dates(cls, portfolio_id: str) -> Tuple[dt.date, ...]:
        position_dates = cls.get_session()._get('/portfolios/{id}/positions/dates'.format(id=portfolio_id))['results']
        return tuple(dt.datetime.strptime(d, '%Y-%m-%d').date() for d in position_dates)

    @classmethod
    @backoff.on_exception(lambda: backoff.expo(base=2, factor=2),
                          (MqTimeoutError, MqInternalServerError),
                          max_tries=5)
    @backoff.on_exception(lambda: backoff.constant(90),
                          MqRateLimitedError,
                          max_tries=5)
    def update_positions(cls,
                         portfolio_id: str,
                         position_sets: List[PositionSet],
                         net_positions: bool = True) -> float:
        url = f'/portfolios/{portfolio_id}/positions?netPositions={str(net_positions).lower()}'
        return GsSession.current._put(url, position_sets)

    @classmethod
    @backoff.on_exception(lambda: backoff.expo(base=2, factor=2),
                          (MqTimeoutError, MqInternalServerError),
                          max_tries=5)
    @backoff.on_exception(lambda: backoff.constant(90),
                          MqRateLimitedError,
                          max_tries=5)
    def get_positions_data(cls,
                           portfolio_id: str,
                           start_date: dt.date,
                           end_date: dt.date,
                           fields: List[str] = None,
                           performance_report_id: str = None,
                           position_type: PositionType = None,
                           include_all_business_days: bool = False) -> List[dict]:
        start_date_str = start_date.isoformat()
        end_date_str = end_date.isoformat()
        url = f'/portfolios/{portfolio_id}/positions/data?startDate={start_date_str}&endDate={end_date_str}'
        if fields is not None:
            url += '&fields='.join([''] + fields)
        if performance_report_id is not None:
            url += f'&reportId={performance_report_id}'
        if position_type is not None:
            url += '&type=' + position_type.value
        if include_all_business_days:
            url += '&includeAllBusinessDays=true'

        return GsSession.current._get(url)['results']

    @classmethod
    def update_quote(cls, quote_id: str, request: RiskRequest):
        return cls.get_session()._put('/risk-internal/quote/save/{id}'.format(id=quote_id), request)

    @classmethod
    def save_quote(cls, request: RiskRequest) -> str:
        return cls.get_session()._post('/risk-internal/quote/save', request)['results']

    @classmethod
    def update_workflow_quote(cls, quote_id: str, request: SaveQuoteRequest):
        headers = {'Content-Type': 'application/x-msgpack'}
        return cls.get_session()._put('/risk-internal/quote/workflow/save/{id}'.format(id=quote_id), tuple([request]),
                                      request_headers=headers)['results']

    @classmethod
    def save_workflow_quote(cls, request: SaveQuoteRequest) -> str:
        headers = {'Content-Type': 'application/x-msgpack'}
        return cls.get_session()._post('/risk-internal/quote/workflow/save', tuple([request]),
                                       request_headers=headers)['results']

    @classmethod
    def share_workflow_quote(cls, request: SaveQuoteRequest) -> str:
        headers = {'Content-Type': 'application/x-msgpack'}
        return cls.get_session()._post('/risk-internal/quote/workflow/share', tuple([request]),
                                       request_headers=headers)['results']

    @classmethod
    def get_workflow_quote(cls, workflow_id: str) -> Tuple[WorkflowPosition, ...]:
        url = f'/risk-internal/quote/workflow/{workflow_id}'
        results = cls.get_session()._get(url, timeout=181)
        wf_pos_res = WorkflowPositionsResponse.from_dict(results)
        if wf_pos_res:
            return wf_pos_res.results
        else:
            return ()

    @classmethod
    def get_shared_workflow_quote(cls, workflow_id: str) -> Tuple[WorkflowPosition, ...]:
        url = f'/risk-internal/quote/workflow/shared/{workflow_id}'
        results = cls.get_session()._get(url, timeout=181)
        wf_pos_res = WorkflowPositionsResponse.from_dict(results)
        if wf_pos_res:
            return wf_pos_res.results
        else:
            return ()

    @classmethod
    def save_to_shadowbook(cls, request: RiskRequest, name: str) -> str:
        return cls.get_session()._put(f'/risk-internal/shadowbook/save/{name}', request)['results']

    @classmethod
    def get_risk_models_by_coverage(cls, portfolio_id: str, term: Term = Term.Medium):
        return cls.get_session()._get(f'/portfolios/{portfolio_id}/models?sortByTerm={term.value}')['results']

    @classmethod
    @backoff.on_exception(lambda: backoff.expo(base=2, factor=2),
                          (MqTimeoutError, MqInternalServerError),
                          max_tries=5)
    @backoff.on_exception(lambda: backoff.constant(90),
                          MqRateLimitedError,
                          max_tries=5)
    def get_reports(cls, portfolio_id: str, tags: Dict) -> Tuple[Report, ...]:
        results = cls.get_session()._get('/portfolios/{id}/reports'.format(id=portfolio_id), cls=Report)['results']
        if tags is not None:
            tags_as_list = [{'name': key, 'value': tags[key]} for key in tags]
            results = [r for r in results if r.parameters.tags == tags_as_list]
        return results

    @classmethod
    @backoff.on_exception(lambda: backoff.expo(base=2, factor=2),
                          (MqTimeoutError, MqInternalServerError),
                          max_tries=5)
    @backoff.on_exception(lambda: backoff.constant(90),
                          MqRateLimitedError,
                          max_tries=5)
    def schedule_reports(cls,
                         portfolio_id: str,
                         start_date: dt.date = None,
                         end_date: dt.date = None,
                         backcast: bool = False) -> dict:
        payload = {'parameters': {'backcast': backcast}}
        if start_date is not None:
            payload['startDate'] = start_date.isoformat()
        if end_date is not None:
            payload['endDate'] = end_date.isoformat()
        portfolio = cls.get_portfolio(portfolio_id)
        if portfolio.tag_name_hierarchy is None or len(portfolio.tag_name_hierarchy) == 0:
            cls.get_session()._post(f'/portfolios/{portfolio_id}/schedule', payload)
        else:
            count = 10
            for report_id in portfolio.report_ids:
                if count == 0:
                    sleep(2)
                    count = 10
                    cls.get_session()._post(f'/reports/{report_id}/schedule', payload)
                else:
                    cls.get_session()._post(f'/reports/{report_id}/schedule', payload)
                    count -= 1

    @classmethod
    @backoff.on_exception(lambda: backoff.expo(base=2, factor=2),
                          (MqTimeoutError, MqInternalServerError),
                          max_tries=5)
    @backoff.on_exception(lambda: backoff.constant(90),
                          MqRateLimitedError,
                          max_tries=5)
    def get_schedule_dates(cls,
                           portfolio_id: str,
                           backcast: bool = False) -> List[dt.date]:
        results = cls.get_session()._get(f'/portfolios/{portfolio_id}/schedule/dates?backcast={str(backcast).lower()}')
        return [dt.datetime.strptime(results['startDate'], '%Y-%m-%d').date(),
                dt.datetime.strptime(results['endDate'], '%Y-%m-%d').date()]

    @classmethod
    @deprecation.deprecated(deprecated_in='1.0.10',
                            details='GsPortfolioApi.get_custom_aum is now deprecated, please use '
                                    'GsReportApi.get_custom_aum instead.')
    def get_custom_aum(cls,
                       portfolio_id: str,
                       start_date: dt.date = None,
                       end_date: dt.date = None) -> dict:
        url = f'/portfolios/{portfolio_id}/aum?'
        if start_date:
            url += f"&startDate={start_date.strftime('%Y-%m-%d')}"
        if end_date:
            url += f"&endDate={end_date.strftime('%Y-%m-%d')}"
        return GsSession.current._get(url)['data']

    @classmethod
    @deprecation.deprecated(deprecated_in='1.0.10',
                            details='GsPortfolioApi.upload_custom_aum is now deprecated, please use '
                                    'GsReportApi.upload_custom_aum instead.')
    def upload_custom_aum(cls,
                          portfolio_id: str,
                          aum_data: List[Dict],
                          clear_existing_data: bool = None) -> dict:
        url = f'/portfolios/{portfolio_id}/aum'
        payload = {'data': aum_data}
        if clear_existing_data:
            url += '?clearExistingData=true'
        return GsSession.current._post(url, payload)

    @classmethod
    def update_portfolio_tree(cls, portfolio_id: str):
        return GsSession.current._post(f'/portfolios/{portfolio_id}/tree', {})

    @classmethod
    def get_portfolio_tree(cls, portfolio_id: str):
        return GsSession.current._get(f'/portfolios/{portfolio_id}/tree', cls=PortfolioTree)

    @classmethod
    def get_attribution(cls,
                        portfolio_id: str,
                        start_date: dt.date = None,
                        end_date: dt.date = None,
                        currency: Currency = None,
                        performance_report_id: str = None) -> Dict:
        url = f'/attribution/{portfolio_id}?'
        if start_date:
            url += f"&startDate={start_date.strftime('%Y-%m-%d')}"
        if end_date:
            url += f"&endDate={end_date.strftime('%Y-%m-%d')}"
        if currency:
            url += f"&currency={currency.value}"
        if performance_report_id:
            url += f'&reportId={performance_report_id}'
        return GsSession.current._get(url)['results']



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import logging
from typing import Iterable, Union, Tuple

import gs_quant.target.backtests as backtests
from gs_quant.api.gs.backtests import GsBacktestApi
from gs_quant.api.gs.backtests_xasset.apis import GsBacktestXassetApi
from gs_quant.api.gs.backtests_xasset.request import BasicBacktestRequest
from gs_quant.api.gs.backtests_xasset.response_datatypes.backtest_datatypes import DateConfig, Trade, Configuration, \
    RollDateMode, TransactionCostConfig
from gs_quant.backtests.core import Backtest, TradeInMethod
from gs_quant.base import get_enum_value, Base
from gs_quant.common import Currency
from gs_quant.common import FieldValueMap, AssetClass
from gs_quant.errors import MqValueError
from gs_quant.instrument import EqOption, EqVarianceSwap, FXOption, FXBinary, FXCorrelationSwap, FXOneTouch, \
    FXVarianceSwap, FXVolatilitySwap, Instrument, IRSwaption
from gs_quant.target.backtests import BacktestResult, BacktestRisk, \
    BacktestTradingQuantityType, DeltaHedgeParameters, \
    BacktestSignalSeriesItem, \
    BacktestStrategyUnderlier, BacktestStrategyUnderlierHedge, EquityMarketModel, BacktestTradingParameters, \
    FlowVolBacktestMeasure

_logger = logging.getLogger(__name__)

BACKTEST_TYPE_NAME = 'VolatilityFlow'
BACKTEST_TYPE_VALUE = 'Volatility Flow'
ISO_FORMAT = r"^([0-9]{4})-([0-9]{2})-([0-9]{2})$"


class StrategySystematic:
    """Equity back testing systematic strategy"""
    _supported_eq_instruments = (EqOption, EqVarianceSwap)
    _supported_fx_instruments = (FXOption, FXBinary, FXCorrelationSwap, FXOneTouch, FXVarianceSwap, FXVolatilitySwap)
    _supported_ir_instruments = (IRSwaption,)
    _supported_instruments = _supported_eq_instruments + _supported_fx_instruments + _supported_ir_instruments

    def __init__(self,
                 underliers: Union[Instrument, Iterable[Instrument]],
                 quantity: float = 1,
                 quantity_type: Union[BacktestTradingQuantityType, str] = BacktestTradingQuantityType.notional,
                 trade_in_method: Union[TradeInMethod, str] = TradeInMethod.FixedRoll,
                 roll_frequency: str = None,
                 scaling_method: str = None,
                 index_initial_value: float = 0.0,
                 delta_hedge: DeltaHedgeParameters = None,
                 name: str = None,
                 cost_netting: bool = False,
                 currency: Union[Currency, str] = Currency.USD,
                 trade_in_signals: Tuple[BacktestSignalSeriesItem, ...] = None,
                 trade_out_signals: Tuple[BacktestSignalSeriesItem, ...] = None,
                 market_model: Union[EquityMarketModel, str] = EquityMarketModel.SFK,
                 roll_date_mode: str = None,
                 expiry_date_mode: str = None,
                 cash_accrual: bool = True,
                 combine_roll_signal_entries: bool = False,
                 transaction_cost_config: TransactionCostConfig = None,
                 use_xasset_backtesting_service: bool = False):
        self.__cost_netting = cost_netting
        self.__currency = get_enum_value(Currency, currency)
        self.__name = name
        self.__backtest_type = BACKTEST_TYPE_NAME
        self.__cash_accrual = cash_accrual

        trade_in_method = get_enum_value(TradeInMethod, trade_in_method).value
        market_model = get_enum_value(EquityMarketModel, market_model).value

        self.__trading_parameters = BacktestTradingParameters(
            quantity=quantity,
            quantity_type=get_enum_value(BacktestTradingQuantityType, quantity_type).value,
            trade_in_method=trade_in_method,
            roll_frequency=roll_frequency,
            trade_in_signals=trade_in_signals,
            trade_out_signals=trade_out_signals,
            roll_date_mode=roll_date_mode
        )

        self.__underliers = []
        trade_instruments = []
        if isinstance(underliers, self._supported_instruments):
            instrument = underliers
            notional_percentage = 100
            trade_instruments.append(instrument)
            self.__underliers.append(BacktestStrategyUnderlier(
                instrument=instrument,
                notional_percentage=notional_percentage,
                hedge=BacktestStrategyUnderlierHedge(risk_details=delta_hedge),
                market_model=market_model,
                expiry_date_mode=expiry_date_mode))
        else:
            for underlier in underliers:
                if isinstance(underlier, tuple):
                    instrument = underlier[0]
                    notional_percentage = underlier[1]
                else:
                    instrument = underlier
                    notional_percentage = 100

                if not isinstance(instrument, self._supported_instruments):
                    raise MqValueError('The format of the backtest asset is incorrect.')
                else:
                    instrument = instrument.scale(notional_percentage / 100, in_place=False, check_resolved=False)
                trade_instruments.append(instrument)
                self.__underliers.append(BacktestStrategyUnderlier(
                    instrument=instrument,
                    notional_percentage=notional_percentage,
                    hedge=BacktestStrategyUnderlierHedge(risk_details=delta_hedge),
                    market_model=market_model,
                    expiry_date_mode=expiry_date_mode))
        # xasset backtesting service fields
        trade_buy_dates = tuple(s.date for s in trade_in_signals if s.value) if trade_in_signals is not None else None
        trade_exit_dates = tuple(s.date for s in trade_out_signals if s.value) if trade_out_signals is not None else \
            None
        self.__trades = (Trade(tuple(trade_instruments), roll_frequency, trade_buy_dates, roll_frequency,
                               trade_exit_dates, quantity, quantity_type),)
        self.__delta_hedge_frequency = '1b' if delta_hedge else None
        self.__transaction_cost_config = transaction_cost_config
        self.__xasset_bt_service_config = Configuration(roll_date_mode=RollDateMode(roll_date_mode) if
                                                        roll_date_mode is not None else None,
                                                        market_model=EquityMarketModel(market_model) if
                                                        market_model else None,
                                                        cash_accrual=cash_accrual,
                                                        combine_roll_signal_entries=combine_roll_signal_entries)

        backtest_parameters_class: Base = getattr(backtests, self.__backtest_type + 'BacktestParameters')
        backtest_parameter_args = {
            'trading_parameters': self.__trading_parameters,
            'underliers': self.__underliers,
            'trade_in_method': trade_in_method,
            'scaling_method': scaling_method,
            'index_initial_value': index_initial_value
        }
        self.__backtest_parameters = backtest_parameters_class.from_dict(backtest_parameter_args)
        all_eq = all(isinstance(i, self._supported_eq_instruments) for i in trade_instruments)
        all_fx = all(isinstance(i, self._supported_fx_instruments) for i in trade_instruments)
        all_ir = all(isinstance(i, self._supported_ir_instruments) for i in trade_instruments)
        if not (all_eq or all_fx or all_ir):
            raise MqValueError('Cannot run backtests for different asset classes.')
        self.__use_xasset_backtesting_service = all_fx or all_ir or use_xasset_backtesting_service

    def __run_service_based_backtest(self, start: dt.date, end: dt.date,
                                     measures: Iterable[FlowVolBacktestMeasure]) -> BacktestResult:
        date_cfg = DateConfig(start, end)
        if not measures:
            measures = (FlowVolBacktestMeasure.PNL,)
        basic_bt_request = BasicBacktestRequest(date_cfg, self.__trades, measures, self.__delta_hedge_frequency,
                                                self.__transaction_cost_config, self.__xasset_bt_service_config)
        basic_bt_response = GsBacktestXassetApi.calculate_basic_backtest(basic_bt_request, decode_instruments=False)
        risks = tuple(
            BacktestRisk(name=k.value,
                         timeseries=tuple(FieldValueMap(date=d, value=r.result) for d, r in v.items()))
            for k, v in basic_bt_response.measures.items()
        )
        portfolio = []
        for d in sorted(set().union(basic_bt_response.portfolio.keys(), basic_bt_response.transactions.keys())):
            if d in basic_bt_response.portfolio:
                positions = [{'instrument': i if i is not None else {}, 'quantity': self.__position_quantity(i)} for
                             i in basic_bt_response.portfolio[d]]
            else:
                positions = []
            transactions = []
            if d in basic_bt_response.transactions:
                for t in basic_bt_response.transactions[d]:
                    trades = [{'instrument': i if i is not None else {},
                               'price': t.portfolio_price,
                               'quantity': t.quantity if t.quantity is not None else None}
                              for i in t.portfolio] if t.portfolio is not None else []
                    transactions.append({'type': t.direction.value, 'trades': trades, 'cost': t.cost})
            portfolio.append({'date': d, 'positions': positions, 'transactions': transactions})
        return BacktestResult(risks=risks, portfolio=portfolio)

    def __position_quantity(self, instrument: dict) -> float:
        if instrument['assetClass'] == AssetClass.Equity.value:
            direction = 1 if instrument['buySell'] == 'Buy' else -1
            quantity = instrument['numberOfOptions'] if instrument['type'] == 'Option' else instrument['quantity']
            return direction * quantity
        return None

    def backtest(
            self,
            start: dt.date = None,
            end: dt.date = dt.date.today() - dt.timedelta(days=1),
            is_async: bool = False,
            measures: Iterable[FlowVolBacktestMeasure] = (FlowVolBacktestMeasure.ALL_MEASURES,),
            correlation_id: str = None
    ) -> Union[Backtest, BacktestResult]:
        if self.__use_xasset_backtesting_service:
            return self.__run_service_based_backtest(start, end, measures)
        params_dict = self.__backtest_parameters.as_dict()
        params_dict['measures'] = [m.value for m in measures]
        backtest_parameters_class: Base = getattr(backtests, self.__backtest_type + 'BacktestParameters')
        params = backtest_parameters_class.from_dict(params_dict)

        backtest = Backtest(name=self.__name,
                            mq_symbol=self.__name,
                            parameters=params,
                            start_date=start,
                            end_date=end,
                            type=BACKTEST_TYPE_VALUE,
                            asset_class=AssetClass.Equity,
                            currency=self.__currency,
                            cost_netting=self.__cost_netting,
                            cash_accrual=self.__cash_accrual)

        if is_async:
            # Create back test ...
            response = GsBacktestApi.create_backtest(backtest)

            # ... and schedule it
            GsBacktestApi.schedule_backtest(backtest_id=response.id)
        else:
            # Run on-the-fly back test
            response = GsBacktestApi.run_backtest(backtest, correlation_id)

        return response



"""
Copyright 2023 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""

import unittest
import datetime as dt
from gs_quant.data.utilities import SecmasterXrefFormatter


class TestSecmasterXrefFormatter(unittest.TestCase):

    def setUp(self):
        """Set up test fixtures before each test method."""
        self.formatter = SecmasterXrefFormatter()

    def test_event_creation_and_priority(self):
        """Test Event dataclass creation and priority assignment."""
        start_event = SecmasterXrefFormatter.Event(
            date="2023-01-01",
            event_type=SecmasterXrefFormatter.EventType.START,
            record={"type": "CUSIP", "value": "12345", "startDate": "2023-01-01", "endDate": "2023-12-31"}
        )

        end_event = SecmasterXrefFormatter.Event(
            date="2023-01-01",
            event_type=SecmasterXrefFormatter.EventType.END,
            record={"type": "ISIN", "value": "US12345", "startDate": "2023-01-01", "endDate": "2023-12-31"}
        )

        self.assertEqual(start_event.priority, 0)
        self.assertEqual(end_event.priority, 1)

    def test_date_sort_key(self):
        """Test date sorting functionality."""
        normal_date = "2023-01-01"
        infinity_date = SecmasterXrefFormatter.INFINITY_DATE

        normal_key = SecmasterXrefFormatter._date_sort_key(normal_date)
        infinity_key = SecmasterXrefFormatter._date_sort_key(infinity_date)

        self.assertEqual(normal_key, dt.datetime(2023, 1, 1))
        self.assertEqual(infinity_key, dt.datetime(9999, 12, 31))
        self.assertTrue(infinity_key > normal_key)

    def test_add_one_day(self):
        """Test adding one day to various date scenarios."""
        # Normal date
        result = SecmasterXrefFormatter._add_one_day("2023-01-01")
        self.assertEqual(result, "2023-01-02")

        # End of month
        result = SecmasterXrefFormatter._add_one_day("2023-01-31")
        self.assertEqual(result, "2023-02-01")

        # End of year
        result = SecmasterXrefFormatter._add_one_day("2023-12-31")
        self.assertEqual(result, "2024-01-01")

        # Leap year
        result = SecmasterXrefFormatter._add_one_day("2024-02-28")
        self.assertEqual(result, "2024-02-29")

        # Infinity date
        result = SecmasterXrefFormatter._add_one_day(SecmasterXrefFormatter.INFINITY_DATE)
        self.assertIsNone(result)

        # Invalid date
        result = SecmasterXrefFormatter._add_one_day("invalid-date")
        self.assertIsNone(result)

    def test_subtract_one_day(self):
        """Test subtracting one day from various date scenarios."""
        # Normal date
        result = SecmasterXrefFormatter._subtract_one_day("2023-01-02")
        self.assertEqual(result, "2023-01-01")

        # Beginning of month
        result = SecmasterXrefFormatter._subtract_one_day("2023-02-01")
        self.assertEqual(result, "2023-01-31")

        # Beginning of year
        result = SecmasterXrefFormatter._subtract_one_day("2024-01-01")
        self.assertEqual(result, "2023-12-31")

        # Invalid date (should return original)
        result = SecmasterXrefFormatter._subtract_one_day("invalid-date")
        self.assertEqual(result, "invalid-date")

    def test_create_events(self):
        """Test event creation from records."""
        records = [
            {
                "type": "CUSIP",
                "value": "12345",
                "startDate": "2023-01-01",
                "endDate": "2023-06-30"
            },
            {
                "type": "ISIN",
                "value": "US12345",
                "startDate": "2023-07-01",
                "endDate": SecmasterXrefFormatter.INFINITY_DATE
            }
        ]

        events = SecmasterXrefFormatter._create_events(records)

        # Should have 3 events: 2 start events + 1 end event (no end event for infinity)
        self.assertEqual(len(events), 3)

        # Check start events
        start_events = [e for e in events if e.event_type == SecmasterXrefFormatter.EventType.START]
        self.assertEqual(len(start_events), 2)

        # Check end events
        end_events = [e for e in events if e.event_type == SecmasterXrefFormatter.EventType.END]
        self.assertEqual(len(end_events), 1)
        self.assertEqual(end_events[0].date, "2023-07-01")  # Day after end date

    def test_convert_empty_data(self):
        """Test conversion with empty data."""
        result = SecmasterXrefFormatter.convert({})
        self.assertEqual(result, {})

        result = SecmasterXrefFormatter.convert({"entity1": []})
        self.assertEqual(result, {"entity1": {"xrefs": []}})

    def test_convert_simple_case(self):
        """Test conversion with a simple single identifier case."""
        data = {
            "entity1": [
                {
                    "type": "CUSIP",
                    "value": "12345",
                    "startDate": "2023-01-01",
                    "endDate": "2023-12-31"
                }
            ]
        }

        result = SecmasterXrefFormatter.convert(data)
        expected = {
            "entity1": {
                "xrefs": [
                    {
                        "startDate": "2023-01-01",
                        "endDate": "2023-12-31",
                        "identifiers": {"CUSIP": "12345"}
                    }
                ]
            }
        }

        self.assertEqual(result, expected)

    def test_convert_infinity_marker_normalization(self):
        """Test that infinity markers are converted to infinity dates."""
        data = {
            "entity1": [
                {
                    "type": "CUSIP",
                    "value": "12345",
                    "startDate": "2023-01-01",
                    "endDate": SecmasterXrefFormatter.INFINITY_MARKER
                }
            ]
        }

        result = SecmasterXrefFormatter.convert(data)
        expected = {
            "entity1": {
                "xrefs": [
                    {
                        "startDate": "2023-01-01",
                        "endDate": SecmasterXrefFormatter.INFINITY_DATE,
                        "identifiers": {"CUSIP": "12345"}
                    }
                ]
            }
        }

        self.assertEqual(result, expected)

    def test_convert_overlapping_periods(self):
        """Test conversion with overlapping identifier periods."""
        data = {
            "entity1": [
                {
                    "type": "CUSIP",
                    "value": "12345",
                    "startDate": "2023-01-01",
                    "endDate": "2023-06-30"
                },
                {
                    "type": "ISIN",
                    "value": "US12345",
                    "startDate": "2023-03-01",
                    "endDate": "2023-09-30"
                }
            ]
        }

        result = SecmasterXrefFormatter.convert(data)
        xrefs = result["entity1"]["xrefs"]

        # Should have 3 periods
        self.assertEqual(len(xrefs), 3)

        # Period 1: Only CUSIP (Jan 1 - Feb 28)
        self.assertEqual(xrefs[0]["startDate"], "2023-01-01")
        self.assertEqual(xrefs[0]["endDate"], "2023-02-28")
        self.assertEqual(xrefs[0]["identifiers"], {"CUSIP": "12345"})

        # Period 2: Both CUSIP and ISIN (Mar 1 - Jun 30)
        self.assertEqual(xrefs[1]["startDate"], "2023-03-01")
        self.assertEqual(xrefs[1]["endDate"], "2023-06-30")
        self.assertEqual(xrefs[1]["identifiers"], {"CUSIP": "12345", "ISIN": "US12345"})

        # Period 3: Only ISIN (Jul 1 - Sep 30)
        self.assertEqual(xrefs[2]["startDate"], "2023-07-01")
        self.assertEqual(xrefs[2]["endDate"], "2023-09-30")
        self.assertEqual(xrefs[2]["identifiers"], {"ISIN": "US12345"})

    def test_convert_adjacent_periods(self):
        """Test conversion with adjacent periods (no gaps)."""
        data = {
            "entity1": [
                {
                    "type": "CUSIP",
                    "value": "12345",
                    "startDate": "2023-01-01",
                    "endDate": "2023-06-30"
                },
                {
                    "type": "CUSIP",
                    "value": "67890",
                    "startDate": "2023-07-01",
                    "endDate": "2023-12-31"
                }
            ]
        }

        result = SecmasterXrefFormatter.convert(data)
        xrefs = result["entity1"]["xrefs"]

        # Should have 2 periods with no gaps
        self.assertEqual(len(xrefs), 2)

        self.assertEqual(xrefs[0]["startDate"], "2023-01-01")
        self.assertEqual(xrefs[0]["endDate"], "2023-06-30")
        self.assertEqual(xrefs[0]["identifiers"], {"CUSIP": "12345"})

        self.assertEqual(xrefs[1]["startDate"], "2023-07-01")
        self.assertEqual(xrefs[1]["endDate"], "2023-12-31")
        self.assertEqual(xrefs[1]["identifiers"], {"CUSIP": "67890"})

    def test_convert_same_date_start_end(self):
        """Test conversion when identifiers start and end on the same date."""
        data = {
            "entity1": [
                {
                    "type": "CUSIP",
                    "value": "12345",
                    "startDate": "2023-01-01",
                    "endDate": "2023-06-30"
                },
                {
                    "type": "ISIN",
                    "value": "US12345",
                    "startDate": "2023-06-30",
                    "endDate": "2023-12-31"
                }
            ]
        }

        result = SecmasterXrefFormatter.convert(data)
        xrefs = result["entity1"]["xrefs"]

        # Should handle the overlap correctly
        self.assertTrue(len(xrefs) >= 1)

    def test_convert_multiple_entities(self):
        """Test conversion with multiple entities."""
        data = {
            "entity1": [
                {
                    "type": "CUSIP",
                    "value": "12345",
                    "startDate": "2023-01-01",
                    "endDate": "2023-12-31"
                }
            ],
            "entity2": [
                {
                    "type": "ISIN",
                    "value": "US67890",
                    "startDate": "2023-01-01",
                    "endDate": "2023-12-31"
                }
            ]
        }

        result = SecmasterXrefFormatter.convert(data)

        self.assertIn("entity1", result)
        self.assertIn("entity2", result)
        self.assertEqual(len(result["entity1"]["xrefs"]), 1)
        self.assertEqual(len(result["entity2"]["xrefs"]), 1)
        self.assertEqual(result["entity1"]["xrefs"][0]["identifiers"], {"CUSIP": "12345"})
        self.assertEqual(result["entity2"]["xrefs"][0]["identifiers"], {"ISIN": "US67890"})

    def test_event_sorting(self):
        """Test that events are sorted correctly by date and priority."""
        events = [
            SecmasterXrefFormatter.Event("2023-01-02", SecmasterXrefFormatter.EventType.START, {}),
            SecmasterXrefFormatter.Event("2023-01-01", SecmasterXrefFormatter.EventType.END, {}),
            SecmasterXrefFormatter.Event("2023-01-01", SecmasterXrefFormatter.EventType.START, {}),
            SecmasterXrefFormatter.Event("2023-01-02", SecmasterXrefFormatter.EventType.END, {}),
        ]

        events.sort(key=lambda e: (SecmasterXrefFormatter._date_sort_key(e.date), e.priority))

        # Should be ordered: 2023-01-01 START, 2023-01-01 END, 2023-01-02 START, 2023-01-02 END
        self.assertEqual(events[0].date, "2023-01-01")
        self.assertEqual(events[0].event_type, SecmasterXrefFormatter.EventType.START)
        self.assertEqual(events[1].date, "2023-01-01")
        self.assertEqual(events[1].event_type, SecmasterXrefFormatter.EventType.END)
        self.assertEqual(events[2].date, "2023-01-02")
        self.assertEqual(events[2].event_type, SecmasterXrefFormatter.EventType.START)
        self.assertEqual(events[3].date, "2023-01-02")
        self.assertEqual(events[3].event_type, SecmasterXrefFormatter.EventType.END)


if __name__ == '__main__':
    # Run the tests
    unittest.main()



"""
Copyright 2019 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
from dataclasses import dataclass, field
from enum import Enum
from typing import ClassVar, List, Optional, Iterable, Union

from dataclasses_json import dataclass_json, config

from gs_quant.backtests.actions import AddTradeAction, AddTradeActionInfo, AddScaledTradeAction, \
    AddScaledTradeActionInfo, HedgeAction, HedgeActionInfo, Action
from gs_quant.backtests.backtest_objects import BackTest, PredefinedAssetBacktest
from gs_quant.backtests.backtest_utils import make_list, CalcType
from gs_quant.backtests.data_sources import DataSource, GsDataSource
from gs_quant.base import field_metadata, exclude_none, static_field
from gs_quant.data import Dataset
from gs_quant.datetime.relative_date import RelativeDateSchedule
from gs_quant.json_convertors import decode_iso_date_or_datetime, decode_date_tuple, dc_decode
from gs_quant.json_convertors_common import encode_risk_measure, decode_risk_measure
from gs_quant.risk import RiskMeasure
from gs_quant.risk.transform import Transformer


class TriggerDirection(Enum):
    ABOVE = 1
    BELOW = 2
    EQUAL = 3


class AggType(Enum):
    ALL_OF = 1
    ANY_OF = 2


@dataclass_json
@dataclass
class TriggerRequirements:

    __sub_classes: ClassVar[List[type]] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        TriggerRequirements.__sub_classes.append(cls)

    @staticmethod
    def sub_classes():
        return tuple(TriggerRequirements.__sub_classes)

    def get_trigger_times(self):
        return []

    @property
    def calc_type(self):
        return CalcType.simple


@dataclass_json
@dataclass
class TriggerInfo(object):
    triggered: bool
    info_dict: Optional[dict] = None

    def __eq__(self, other):
        return self.triggered is other

    def __bool__(self):
        return self.triggered


def check_barrier(direction, test_value, trigger_level) -> TriggerInfo:
    if direction == TriggerDirection.ABOVE:
        if test_value > trigger_level:
            return TriggerInfo(True)
    elif direction == TriggerDirection.BELOW:
        if test_value < trigger_level:
            return TriggerInfo(True)
    else:
        if test_value == trigger_level:
            return TriggerInfo(True)
    return TriggerInfo(False)


@dataclass_json
@dataclass
class PeriodicTriggerRequirements(TriggerRequirements):
    start_date: Optional[dt.date] = field(default=None, metadata=field_metadata)
    end_date: Optional[dt.date] = field(default=None, metadata=field_metadata)
    frequency: Optional[str] = field(default=None, metadata=field_metadata)
    calendar: Optional[Iterable[dt.date]] = field(default=None, metadata=config(exclude=exclude_none,
                                                                                decoder=decode_date_tuple))
    trigger_dates = []
    class_type: str = static_field('periodic_trigger_requirements')

    def get_trigger_times(self) -> [dt.date]:
        if not self.trigger_dates:
            self.trigger_dates = RelativeDateSchedule(self.frequency, self.start_date,
                                                      self.end_date).apply_rule(holiday_calendar=self.calendar)
        return self.trigger_dates

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        if not self.trigger_dates:
            self.get_trigger_times()
        if state in self.trigger_dates:
            next_state = None
            if self.trigger_dates.index(state) != len(self.trigger_dates) - 1:
                next_state = self.trigger_dates[self.trigger_dates.index(state) + 1]
            return TriggerInfo(True, {AddTradeAction: AddTradeActionInfo(scaling=None, next_schedule=next_state),
                                      AddScaledTradeAction: AddScaledTradeActionInfo(next_schedule=next_state),
                                      HedgeAction: HedgeActionInfo(next_schedule=next_state)})
        return TriggerInfo(False)


@dataclass_json
@dataclass
class IntradayTriggerRequirements(TriggerRequirements):
    start_time: Optional[dt.time] = field(default=None, metadata=field_metadata)
    end_time: Optional[dt.time] = field(default=None, metadata=field_metadata)
    frequency: Optional[float] = field(default=None, metadata=field_metadata)
    class_type: str = static_field('intraday_trigger_requirements')

    def __post_init__(self):
        # generate all the trigger times
        self._trigger_times = []
        time = self.start_time
        while time <= self.end_time:
            self._trigger_times.append(time)
            time = (dt.datetime.combine(dt.date.today(), time) + dt.timedelta(minutes=self.frequency)).time()

    def get_trigger_times(self):
        return self._trigger_times

    def has_triggered(self, state: Union[dt.date, dt.datetime], backtest: BackTest = None) -> TriggerInfo:
        return TriggerInfo(state.time() in self._trigger_times)


@dataclass_json
@dataclass
class MktTriggerRequirements(TriggerRequirements):
    data_source: DataSource = field(default=None, metadata=config(decoder=dc_decode(*DataSource.sub_classes(),
                                                                                    allow_missing=True)))
    trigger_level: float = field(default=None, metadata=field_metadata)
    direction: TriggerDirection = field(default=None, metadata=field_metadata)
    class_type: str = static_field('mkt_trigger_requirements')

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        data_value = self.data_source.get_data(state)
        try:
            triggered = check_barrier(self.direction, data_value, self.trigger_level)
        except TypeError:
            raise RuntimeError(f'unable to determine trigger state on {str(state)}, data value was {data_value}')
        return triggered


@dataclass_json
@dataclass
class RiskTriggerRequirements(TriggerRequirements):
    risk: RiskMeasure = field(default=None, metadata=config(decoder=decode_risk_measure, encoder=encode_risk_measure))
    trigger_level: float = field(default=None, metadata=field_metadata)
    direction: TriggerDirection = field(default=None, metadata=field_metadata)
    risk_transformation: Optional[Transformer] = field(default=None, metadata=field_metadata)
    class_type: str = static_field('risk_trigger_requirements')

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        if state not in backtest.results:
            return TriggerInfo(False)
        if self.risk_transformation is None:
            risk_value = backtest.results[state][self.risk].aggregate()
        else:
            risk_value = backtest.results[state][self.risk].transform(
                risk_transformation=self.risk_transformation).aggregate(
                allow_mismatch_risk_keys=True)
        return check_barrier(self.direction, risk_value, self.trigger_level)

    @property
    def calc_type(self):
        return CalcType.path_dependent


@dataclass_json
@dataclass
class AggregateTriggerRequirements(TriggerRequirements):
    triggers: Iterable[TriggerRequirements] = field(default=None, metadata=field_metadata)
    aggregate_type: AggType = field(default=AggType.ALL_OF, metadata=field_metadata)
    class_type: str = static_field('aggregate_trigger_requirements')

    def __setattr__(self, key, value):
        if key == 'triggers':
            if all([isinstance(v, Trigger) for v in value]):
                value = tuple(v.trigger_requirements for v in value)
        super().__setattr__(key, value)

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        info_dict = {}
        if self.aggregate_type == AggType.ALL_OF:
            for trigger in self.triggers:
                t_info = trigger.has_triggered(state, backtest)
                if not t_info:
                    return TriggerInfo(False)
                else:
                    if t_info.info_dict:
                        info_dict.update(t_info.info_dict)
            return TriggerInfo(True, info_dict)
        elif self.aggregate_type == AggType.ANY_OF:
            triggered = False
            for trigger in self.triggers:
                t_info = trigger.has_triggered(state, backtest)
                if t_info:
                    triggered = True
                    if t_info.info_dict:
                        info_dict.update(t_info.info_dict)
            return TriggerInfo(True, info_dict) if triggered else TriggerInfo(False)
        else:
            raise RuntimeError(f'Unrecognised aggregation type: {self.aggregate_type}')

    @property
    def calc_type(self):
        seen_types = set()
        for trigger in self.triggers:
            seen_types.add(trigger.calc_type)

        if CalcType.path_dependent in seen_types:
            return CalcType.path_dependent
        elif CalcType.semi_path_dependent in seen_types:
            return CalcType.semi_path_dependent
        else:
            return CalcType.simple


@dataclass_json
@dataclass
class NotTriggerRequirements(TriggerRequirements):
    trigger: TriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('not_trigger_requirements')

    def __setattr__(self, key, value):
        if key == 'trigger':
            if isinstance(value, Trigger):
                value = value.trigger_requirements
            super().__setattr__(key, value)

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        t_info = self.trigger.has_triggered(state, backtest)
        if t_info:
            return TriggerInfo(False)
        else:
            return TriggerInfo(True)


@dataclass_json
@dataclass
class DateTriggerRequirements(TriggerRequirements):
    dates: Iterable[Union[dt.datetime, dt.date]] = field(default=None, metadata=config(
        exclude=exclude_none, decoder=decode_iso_date_or_datetime))
    entire_day: bool = field(default=False, metadata=field_metadata)
    class_type: str = static_field('date_trigger_requirements')
    dates_from_datetimes = []

    def __post_init__(self):
        self.dates_from_datetimes = [d.date() if isinstance(d, dt.datetime)
                                     else d for d in self.dates] if self.entire_day else None

    def has_triggered(self, state: Union[dt.date, dt.datetime], backtest: BackTest = None) -> TriggerInfo:
        if self.entire_day:
            dates = sorted(self.dates_from_datetimes)
            if isinstance(state, dt.datetime):
                state = state.date()
        else:
            dates = sorted(self.dates)
        if state in dates:
            next_state = None
            if dates.index(state) < len(dates) - 1:
                next_state = dates[dates.index(state) + 1]
            return TriggerInfo(True, {AddTradeAction: AddTradeActionInfo(scaling=None, next_schedule=next_state),
                                      AddScaledTradeAction: AddScaledTradeActionInfo(next_schedule=next_state),
                                      HedgeAction: HedgeActionInfo(next_schedule=next_state)})
        return TriggerInfo(False)

    def get_trigger_times(self):
        return self.dates_from_datetimes or self.dates


@dataclass_json
@dataclass
class PortfolioTriggerRequirements(TriggerRequirements):
    data_source: str = field(default=None, metadata=field_metadata)
    trigger_level: float = field(default=None, metadata=field_metadata)
    direction: TriggerDirection = field(default=None, metadata=field_metadata)
    class_type: str = static_field('portfolio_trigger_requirements')

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        if self.data_source == 'len':
            value = len(backtest.portfolio_dict)
            if self.direction == TriggerDirection.ABOVE:
                if value > self.trigger_level:
                    return TriggerInfo(True)
            elif self.direction == TriggerDirection.BELOW:
                if value < self.trigger_level:
                    return TriggerInfo(True)
            else:
                if value == self.trigger_level:
                    return TriggerInfo(True)
        return TriggerInfo(False)


@dataclass_json
@dataclass
class MeanReversionTriggerRequirements(TriggerRequirements):
    data_source: DataSource = field(default=None, metadata=config(decoder=dc_decode(*DataSource.sub_classes(),
                                                                                    allow_missing=True)))
    z_score_bound: float = field(default=None, metadata=field_metadata)
    rolling_mean_window: int = field(default=None, metadata=field_metadata)
    rolling_std_window: int = field(default=None, metadata=field_metadata)
    current_position = 0
    class_type: str = static_field('mean_reversion_trigger_requirements')

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        rolling_mean = self.data_source.get_data_range(state, self.rolling_mean_window).mean()
        rolling_std = self.data_source.get_data_range(state, self.rolling_std_window).std()
        current_price = self.data_source.get_data(state)
        if self.current_position == 0:
            if abs((current_price - rolling_mean) / rolling_std) > self.z_score_bound:
                if current_price > rolling_mean:
                    self.current_position = -1
                    return TriggerInfo(True, {AddTradeAction: AddTradeActionInfo(scaling=-1)})
                else:
                    self.current_position = 1
                    return TriggerInfo(True, {AddTradeAction: AddTradeActionInfo(scaling=1)})
        elif self.current_position == 1:
            if current_price > rolling_mean:
                self._current_position = 0
                return TriggerInfo(True, {AddTradeAction: AddTradeActionInfo(scaling=-1)})
        elif self.current_position == -1:
            if current_price > rolling_mean:
                self.current_position = 0
                return TriggerInfo(True, {AddTradeAction: AddTradeActionInfo(scaling=1)})
        else:
            raise RuntimeWarning(f'unexpected current position: {self.current_position}')
        return TriggerInfo(False)


@dataclass_json
@dataclass
class TradeCountTriggerRequirements(TriggerRequirements):
    trade_count: float = field(default=None, metadata=field_metadata)
    direction: TriggerDirection = field(default=None, metadata=field_metadata)
    class_type: str = static_field('trade_count_requirements')

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        value = len(backtest.portfolio_dict.get(state, []))
        if self.direction == TriggerDirection.ABOVE:
            if value > self.trade_count:
                return TriggerInfo(True)
        elif self.direction == TriggerDirection.BELOW:
            if value < self.trade_count:
                return TriggerInfo(True)
        else:
            if value == self.trade_count:
                return TriggerInfo(True)
        return TriggerInfo(False)

    @property
    def calc_type(self):
        return CalcType.path_dependent


@dataclass_json
@dataclass
class EventTriggerRequirements(TriggerRequirements):
    event_name: str = field(default=None, metadata=field_metadata)
    offset_days: int = 0
    data_source: DataSource = field(default=None, metadata=config(decoder=dc_decode(*DataSource.sub_classes(),
                                                                                    allow_missing=True)))
    class_type: str = static_field('event_requirements')
    trigger_dates = []

    def __post_init__(self):
        if self.data_source is None:
            self.data_source = GsDataSource(data_set='MACRO_EVENTS_CALENDAR', asset_id=None, value_header='eventName')

    def get_trigger_times(self) -> [dt.date]:
        if not self.trigger_dates:
            kwargs = {'eventName': self.event_name}
            self.trigger_dates = [d.date() + dt.timedelta(days=self.offset_days) for d in
                                  self.data_source.get_data(None, **kwargs).index]
        return self.trigger_dates

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        dates = sorted(self.trigger_dates)
        if state in dates:
            next_state = None
            if dates.index(state) < len(dates) - 1:
                next_state = dates[dates.index(state) + 1]
            return TriggerInfo(True, {AddTradeAction: AddTradeActionInfo(scaling=None, next_schedule=next_state),
                                      AddScaledTradeAction: AddScaledTradeActionInfo(next_schedule=next_state),
                                      HedgeAction: HedgeActionInfo(next_schedule=next_state)})
        return TriggerInfo(False)

    @staticmethod
    def list_events(currency: str,
                    start=Optional[dt.datetime],
                    end=Optional[dt.datetime], **kwargs):
        kwargs['currency'] = currency
        dataset = Dataset('MACRO_EVENTS_CALENDAR')
        return dataset.get_data(start, end, **kwargs)['eventName'].unique()


@dataclass_json
@dataclass
class Trigger:
    trigger_requirements: Optional[TriggerRequirements] = field(default=None, metadata=field_metadata)
    actions: Union[Action, Iterable[Action]] = field(default=None,
                                                     metadata=config(
                                                         decoder=dc_decode(*Action.sub_classes(), allow_missing=True)))
    __sub_classes: ClassVar[List[type]] = []

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        Trigger.__sub_classes.append(cls)

    @staticmethod
    def sub_classes():
        return tuple(Trigger.__sub_classes)

    def __post_init__(self):
        self.actions = make_list(self.actions)

    def has_triggered(self, state: dt.date, backtest: BackTest = None) -> TriggerInfo:
        """
        implemented by sub classes
        :param state:
        :param backtest:
        :return:
            TriggerInfo containing a bool indication of whether the trigger has triggered and optionally info
            in the form of a dictionary of action type to object info understood by that action.
        """
        return self.trigger_requirements.has_triggered(state, backtest)

    def get_trigger_times(self):
        return self.trigger_requirements.get_trigger_times()

    @property
    def calc_type(self):
        return self.trigger_requirements.calc_type

    @property
    def risks(self):
        return [x.risk for x in make_list(self.actions) if x.risk is not None]


@dataclass_json
@dataclass
class PeriodicTrigger(Trigger):
    trigger_requirements: PeriodicTriggerRequirements = field(default=None, metadata=field_metadata)
    _trigger_dates = None
    class_type: str = static_field('periodic_trigger')


@dataclass_json
@dataclass
class IntradayPeriodicTrigger(Trigger):
    trigger_requirements: IntradayTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('intraday_periodic_trigger')


@dataclass_json
@dataclass
class MktTrigger(Trigger):
    trigger_requirements: MktTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('mkt_trigger')


@dataclass_json
@dataclass
class StrategyRiskTrigger(Trigger):
    trigger_requirements: RiskTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('strategy_risk_trigger')

    @property
    def risks(self):
        return [x.risk for x in make_list(self.actions) if x.risk is not None] + [self.trigger_requirements.risk]


@dataclass_json
@dataclass
class AggregateTrigger(Trigger):
    trigger_requirements: AggregateTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('aggregate_trigger')


@dataclass_json
@dataclass
class NotTrigger(Trigger):
    trigger_requirements: NotTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('not_trigger')


@dataclass_json
@dataclass
class DateTrigger(Trigger):
    trigger_requirements: DateTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('date_trigger')


@dataclass_json
@dataclass
class PortfolioTrigger(Trigger):
    trigger_requirements: PortfolioTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('portfolio_trigger')


@dataclass_json
@dataclass
class MeanReversionTrigger(Trigger):
    trigger_requirements: MeanReversionTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('mean_reversion_trigger')


@dataclass_json
@dataclass
class TradeCountTrigger(Trigger):
    trigger_requirements: TradeCountTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('trade_count_trigger')


@dataclass_json
@dataclass
class EventTrigger(Trigger):
    trigger_requirements: EventTriggerRequirements = field(default=None, metadata=field_metadata)
    class_type: str = static_field('event_trigger')


@dataclass_json
@dataclass
class OrdersGeneratorTrigger(Trigger):
    """Base class for triggers used with the PredefinedAssetEngine."""

    def __post_init__(self):
        if not self.actions:
            self.actions = [Action()]
        super().__post_init__()

    def get_trigger_times(self) -> list:
        """
        Returns the set of times when orders can be generated e.g. every 30 min
        :return: list
        """
        raise RuntimeError('get_trigger_times must be implemented by subclass')

    def generate_orders(self, state: dt.datetime, backtest: PredefinedAssetBacktest = None) -> list:
        """
        Returns the orders generated at state
        :param state: the time when orders are generated
        :param backtest: the backtest, used to access the holdings and orders generated so far
        :return: list
        """
        raise RuntimeError('generate_orders must be implemented by subclass')

    def has_triggered(self, state: dt.datetime, backtest: PredefinedAssetBacktest = None) -> TriggerInfo:
        """
        Calls generate_orders if state is among the trigger times
        :param state: the time of the trigger
        :param backtest: the backtest, used to access the holdings and orders generated so far
        :return: list
        """
        if state.time() not in self.get_trigger_times():
            return TriggerInfo(False)
        else:
            orders = self.generate_orders(state, backtest)
            return TriggerInfo(True, {type(a): orders for a in self.actions}) if len(orders) else TriggerInfo(False)


# These special trigger requirements have a special dependency on other trigger types
triggers_field = AggregateTriggerRequirements.__dataclass_fields__['triggers']
triggers_field.metadata = config(decoder=dc_decode(*TriggerRequirements.sub_classes(), allow_missing=True))

trigger_field = NotTriggerRequirements.__dataclass_fields__['trigger']
trigger_field.metadata = config(decoder=dc_decode(*TriggerRequirements.sub_classes(), allow_missing=True))



"""
Copyright 2020 Goldman Sachs.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing,
software distributed under the License is distributed on an
"AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
KIND, either express or implied.  See the License for the
specific language governing permissions and limitations
under the License.
"""
import datetime as dt
import logging
from typing import Optional, Union

import pandas as pd

import gs_quant.timeseries.measures_reports as ReportMeasures
from gs_quant.api.gs.data import QueryType
from gs_quant.api.gs.portfolios import GsPortfolioApi
from gs_quant.entities.entity import EntityType
from gs_quant.markets.portfolio_manager import PortfolioManager
from gs_quant.timeseries import plot_measure_entity
from gs_quant.timeseries.measures import _extract_series_from_df

LOGGER = logging.getLogger(__name__)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.AUM])
def aum(portfolio_id: str, start_date: dt.date = None, end_date: dt.date = None, *,
        source: str = None, real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Returns the Custom AUM uploaded for the portfolio
    :param portfolio_id: id of portfolio
    :param start_date: start date for getting aum
    :param end_date: end date for getting aum
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio aum values
    """
    data = GsPortfolioApi.get_custom_aum(portfolio_id, start_date, end_date)
    df = pd.DataFrame.from_records(data)
    df = df.set_index(pd.DatetimeIndex(df['date']))
    return _extract_series_from_df(df, QueryType.AUM, True)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.FACTOR_EXPOSURE])
def portfolio_factor_exposure(portfolio_id: str, risk_model_id: str, factor_name: str, unit: str = 'Notional',
                              benchmark_id: str = None, *, source: str = None,
                              real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor exposure data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name
    :param unit: unit in which the timeseries is returned (Notional or Percent)
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor exposure for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.factor_exposure(report.id, factor_name, unit)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.FACTOR_PNL])
def portfolio_factor_pnl(portfolio_id: str, risk_model_id: str, factor_name: str, unit: str = 'Notional',
                         benchmark_id: str = None, *, source: str = None,
                         real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor PnL data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name
    :param unit: unit in which the timeseries is returned (Notional or Percent)
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor pnl for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.factor_pnl(report.id, factor_name, unit)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.FACTOR_PROPORTION_OF_RISK])
def portfolio_factor_proportion_of_risk(portfolio_id: str, risk_model_id: str, factor_name: str,
                                        benchmark_id: str = None, *, source: str = None,
                                        real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Factor proportion of risk data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of factor proportion of risk for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.factor_proportion_of_risk(report.id, factor_name)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.DAILY_RISK])
def portfolio_daily_risk(portfolio_id: str, risk_model_id: str, factor_name: str = 'Total',
                         benchmark_id: str = None, *, source: str = None,
                         real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Daily risk data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name (must be "Factor", "Specific", or "Total")
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily risk for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.daily_risk(report.id, factor_name)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.ANNUAL_RISK])
def portfolio_annual_risk(portfolio_id: str, risk_model_id: str, factor_name: str = 'Total',
                          benchmark_id: str = None, *, source: str = None,
                          real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Annual risk data associated with a factor in a factor risk report

    :param portfolio_id: portfolio id
    :param risk_model_id: factor risk model id
    :param factor_name: factor name (must be "Factor", "Specific", or "Total")
    :param benchmark_id: optional benchmark id
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily risk for requested factor
    """
    pm = PortfolioManager(portfolio_id)
    report = pm.get_factor_risk_report(risk_model_id=risk_model_id, benchmark_id=benchmark_id)
    return ReportMeasures.annual_risk(report.id, factor_name)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.THEMATIC_EXPOSURE])
def portfolio_thematic_exposure(portfolio_id: str, basket_ticker: str, *, source: str = None,
                                real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Thematic exposure of a portfolio to a requested GS thematic flagship basket

    :param portfolio_id: portfolio id
    :param basket_ticker: ticker for thematic basket
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: Timeseries of daily thematic beta of portfolio to requested flagship basket
    """
    pm = PortfolioManager(portfolio_id)
    thematic_report = pm.get_thematic_report()
    return ReportMeasures.thematic_exposure(thematic_report.id, basket_ticker)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.PNL])
def portfolio_pnl(portfolio_id: str, unit: str = 'Notional', *, source: str = None,
                  real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    PNL from all holdings, if unit is Percent, geometrically aggregate over time frame

    :param portfolio_id: id of portfolio
    :param unit: by default uses mode as Notional, but can also be set to Percent
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio pnl
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.pnl(performance_report.id, unit)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.HIT_RATE])
def portfolio_hit_rate(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                       real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Create a time series of the hit rate of a portfolio over a given rolling period

    :param portfolio_id: id of portfolio
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio hit rate
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.hit_rate(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.MAX_DRAWDOWN])
def portfolio_max_drawdown(portfolio_id: str, rolling_window: int, *, source: str = None,
                           real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Create a time series of the max drawdown of a portfolio over a given rolling period

    :param portfolio_id: id of portfolio
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of portfolio hit rate
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.portfolio_max_drawdown(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.MAX_DRAWDOWN])
def portfolio_drawdown_length(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                              real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    The length in days between the peak and the trough of the maximum drawdown

    :param portfolio_id: id of portfolio
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the drawdown length
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.drawdown_length(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.MAX_DRAWDOWN])
def portfolio_max_recovery_period(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                                  real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    The maximum number of days used to reach a previously broken price level over the stated timeframe.

    :param portfolio_id: id of portfolio
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of max recovery period
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.max_recovery_period(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.STANDARD_DEVIATION])
def portfolio_standard_deviation(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                                 real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Volatility of the total return over the stated time frame.

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of standard deviation
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.standard_deviation(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.DOWNSIDE_RISK])
def portfolio_downside_risk(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                            real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Volatility of the periodic returns that are lower than the mean return over the stated time frame.
    Standard deviation is calculated using all returns, downside risk is calculated using only the returns
    that are below the mean.

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of downside_risk
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.downside_risk(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.DOWNSIDE_RISK])
def portfolio_semi_variance(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                            real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Volatility of the periodic returns that are lower than the mean return over the stated time frame.
    Standard deviation is calculated using all returns, semi variance is calculated using only the returns
    that are below 0.

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of downside_risk
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.semi_variance(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.KURTOSIS])
def portfolio_kurtosis(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                       real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Kurtosis measure the peakedness or flatness of the return distribution over the state time frame.
    In a flat distribution the average value is more likely to occur.

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of kurtosis
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.kurtosis(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.SKEWNESS])
def portfolio_skewness(portfolio_id: str, rolling_window: Union[int, str], *, source: str = None,
                       real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Skewness measures the degree of symmetry of the return distribution over the stated time frame.
    If the left tail is more pronounced than the right tail, it is said the return have negative skewness.

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of skewness
    """
    pm = PortfolioManager(portfolio_id)
    performance_report = pm.get_performance_report()
    return ReportMeasures.skewness(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.REALIZED_VAR])
def portfolio_realized_var(portfolio_id: str, rolling_window: Union[int, str], confidence_interval: float = .95,
                           *, source: str = None, real_time: bool = False,
                           request_id: Optional[str] = None) -> pd.Series:
    """
    The maximum expected loss of the portfolios, calculated using
    the natural distribution of returns over a stated time frame and is based on a confidence level.

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param confidence_interval: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of realized var
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.realized_var(performance_report.id, rolling_window, confidence_interval)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.TRACKING_ERROR])
def portfolio_tracking_error(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                             source: str = None,
                             real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    The standard deviation of the excess return relative to the benchmark over the state time frame.

    Tracking Error = sqrt( sum((ExcessRi - ExcessRmean)^2) / (N-1) )

    Where:

    ExcessRi = excess return in period i

    ExcessRmean = mean excess return during bull markets

    N = number of periods

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param benchmark_id: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of tracking error
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.tracking_error(performance_report.id, benchmark_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.TRACKING_ERROR_BEAR])
def portfolio_tracking_error_bear(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                                  source: str = None,
                                  real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    The standard deviation of the excess return relative to the benchmark over the state time frame,
    only counting the periods when benchmark returns were negative.

    Bear Tracking Error = sqrt( sum((ExcessRi - ExcessRmeanBear)^2) / (N-1) ) for all when Benchmark Return < 0

    Where:

    ExcessRi = excess return in period i

    ExcessRmeanBear = mean excess return during bull markets

    N = number of periods

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param benchmark_id: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of tracking error
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.tracking_error_bear(performance_report.id, benchmark_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.TRACKING_ERROR_BULL])
def portfolio_tracking_error_bull(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                                  source: str = None,
                                  real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    The standard deviation of the excess return relative to the benchmark over the state time frame,
    only counting the periods when benchmark returns were positive.

    Bull Tracking Error = sqrt( sum((ExcessRi - ExcessRmeanBull)^2) / (N-1) ) for all when Benchmark Return > 0

    Where:

    ExcessRi = excess return in period i

    ExcessRmeanBull = mean excess return during bull markets

    N = number of periods

    :param portfolio_id: id of portfolio
    :param rolling_window: the rolling window to look back on
    :param benchmark_id: the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of tracking error
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.tracking_error_bull(performance_report.id, benchmark_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.SHARPE_RATIO])
def portfolio_sharpe_ratio(portfolio_id: str, risk_free_id: str, rolling_window: Union[int, str], *,
                           source: str = None,
                           real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the risk free rate per unit of volatility

    Sharpe Ratio = (Rp - Rf) / σp

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    σp = Standard Deviation of Portfolio Return.

    :param portfolio_id: id of portfolio
    :param risk_free_id: id of risk-free asset
    :param rolling_window: size of the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of sharpe ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.portfolio_sharpe_ratio(performance_report.id, risk_free_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.CALMAR_RATIO])
def portfolio_calmar_ratio(portfolio_id: str, rolling_window: Union[int, str], *,
                           source: str = None,
                           real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted performance metric that divides an investment strategy’s
    annualized rate of return by its maximum drawdown, focusing specifically
    on the strategy’s return per unit of downside risk.

    Calmar Ratio = Annualized Return / Max Drawdown

    :param portfolio_id: id of portfolio
    :param rolling_window: size of the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of calmar ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.calmar_ratio(performance_report.id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.SORTINO_RATIO])
def portfolio_sortino_ratio(portfolio_id: str, comparison_id: str, rolling_window: Union[int, str], *,
                            source: str = None,
                            real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark per unit of semi-variance

    Sortino Ratio = (Rp - Rb) / Downside Deviation

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Downside Deviation = Standard Deviation of negative asset returns.

    :param portfolio_id: id of portfolio
    :param comparison_id: id of security to compare to
    :param rolling_window: size of the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the sortino ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.sortino_ratio(performance_report.id, comparison_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.INFORMATION_RATIO])
def portfolio_information_ratio(portfolio_id: str, benchmark_id: str, *,
                                source: str = None,
                                real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark,
    per unit of tracking error volatility. It measures the consistency with
    which the portfolio is beating the benchmark.

    Information Ratio = (Rp - Rb) / Tracking Error

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Tracking Error = Standard Deviation of the difference between the portfolio and benchmark returns.

    :param portfolio_id: id of portfolio
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the information ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.information_ratio(performance_report.id, benchmark_id)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.INFORMATION_RATIO])
def portfolio_information_ratio_bull(portfolio_id: str, benchmark_id: str, *,
                                     source: str = None,
                                     real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark,
    per unit of tracking error volatility, counting only the periods when the benchmark return was positive.

    Information Ratio = (Rp - Rb) / Tracking Error

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Tracking Error = Standard Deviation of the difference between the portfolio and benchmark returns.

    :param portfolio_id: id of portfolio
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the information ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.information_ratio_bull(performance_report.id, benchmark_id)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.INFORMATION_RATIO])
def portfolio_information_ratio_bear(portfolio_id: str, benchmark_id: str, *,
                                     source: str = None,
                                     real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the benchmark,
    per unit of tracking error volatility, counting only the periods when the benchmark return was negative


    Information Ratio = (Rp - Rb) / Tracking Error

    Where:

    Rp = Portfolio Return

    Rb = Benchmark Return

    Tracking Error = Standard Deviation of the difference between the portfolio and benchmark returns.


    :param portfolio_id: id of portfolio
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the information ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.information_ratio_bear(performance_report.id, benchmark_id)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.MODIGLIANI_RATIO])
def portfolio_modigliani_ratio(portfolio_id: str, benchmark_id: str, risk_free_id: str,
                               rolling_window: Union[int, str], *,
                               source: str = None,
                               real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Modigliani RAP measures how much the portfolio would have returns if it had had
    the same risk as the benchmark. It is a linear transformation of the Sharpe Ratio,
    but the results are expressed in terms of performance.

    M2 = (Sharpe Ratio * σb) + Rf

    Where:

    Sharpe Ratio = Portfolio Sharpe Ratio

    σb = Standard Deviation of the Benchmark Return

    Rf = Risk Free Rate.

    :param portfolio_id: id of portfolio
    :param benchmark_id: id of benchmark asset
    :param risk_free_id: id of risk free asset
    :param rolling_window: size of window to use
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the modigliani ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.modigliani_ratio(performance_report.id, benchmark_id, risk_free_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.TREYNOR_RATIO])
def portfolio_treynor_measure(portfolio_id: str, risk_free_id: str, benchmark_id: str, *,
                              source: str = None,
                              real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-adjusted measure that calculates the excess return over the risk free rate
    per unit of beta relative to the benchmark. This is useful for assessing the excess
    return from each unit of systematic risk.

    Treynor Ratio = (Rp - Rf) / Beta

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    :param portfolio_id: id of portfolio
    :param risk_free_id: id of risk-free asset
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the treynor measure
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.treynor_measure(performance_report.id, risk_free_id, benchmark_id)


def portfolio_jensen_alpha(portfolio_id: str, benchmark_id: str, risk_free_id: str, *,
                           source: str = None,
                           real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-Adjusted measure that calculates the actual return of the portfolio
    over and above the return predicted by the CAPM, given the portfolios beta
    and benchmark returns.

    Jensen's Alpha = Rp - [Rf + Beta * (Rb - Rf)]

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    Rb = Benchmark Return

    :param portfolio_id: id of portfolio
    :param risk_free_id: id of risk-free asset
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of jensen alpha
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.jensen_alpha(performance_report.id, benchmark_id, risk_free_id)


def portfolio_jensen_alpha_bear(portfolio_id: str, benchmark_id: str, risk_free_id: str, *,
                                source: str = None,
                                real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-Adjusted measure that calculates the actual return of the portfolio
    over and above the return predicted by the CAPM, given the portfolios beta
    and benchmark returns. calculated based only on the periods when the benchmark return was negative

    Jensen's Alpha = Rp - [Rf + Beta * (Rb - Rf)]

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    Rb = Benchmark Return

    :param portfolio_id: id of portfolio
    :param risk_free_id: id of risk-free asset
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of jensen alpha
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.jensen_alpha_bear(performance_report.id, benchmark_id, risk_free_id)


def portfolio_jensen_alpha_bull(portfolio_id: str, benchmark_id: str, risk_free_id: str, *,
                                source: str = None,
                                real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Risk-Adjusted measure that calculates the actual return of the portfolio
    over and above the return predicted by the CAPM, given the portfolios beta
    and benchmark returns. calculated based only on the periods when the benchmark return was positive

    Jensen's Alpha = Rp - [Rf + Beta * (Rb - Rf)]

    Where:

    Rp = Portfolio Return

    Rf = Risk-Free Rate

    Beta = Portfolio Beta

    Rb = Benchmark Return

    :param portfolio_id: id of portfolio
    :param risk_free_id: id of risk-free asset
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of jensen alpha
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.jensen_alpha_bull(performance_report.id, benchmark_id, risk_free_id)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.ALPHA])
def portfolio_alpha(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                    source: str = None,
                    real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    This is the intercept from a regression analysis, where the portfolio's returns are regressed
    against the benchmark's returns. It represents the portion of the portfolio's return that is
    not explained by the benchmark's performance

    :param portfolio_id: id of portfolio
    :param rolling_window: size of the rolling window to look back on
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the beta
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.alpha(performance_report.id, benchmark_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.BETA])
def portfolio_beta(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                   source: str = None,
                   real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Beta defined as the slope of the regression between report pnl and benchmark pnl

    :param portfolio_id: id of portfolio
    :param rolling_window: size of the rolling window to look back on
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the beta
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.portfolio_beta(performance_report.id, benchmark_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.CORRELATION])
def portfolio_correlation(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                          source: str = None,
                          real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    Correlation coefficient between the portfolio and the benchmark returns over the stated time frame

    :param portfolio_id: id of portfolio
    :param rolling_window: size of the rolling window to look back on
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: time series of the correlation coefficient
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.portfolio_correlation(performance_report.id, benchmark_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.R_SQUARED])
def portfolio_r_squared(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                        source: str = None,
                        real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    A measure of how well the portfolio's performance correlates with the performance of the benchmark,
    and thus a measure of what portion of its performance may be explained by the performance of the benchmark.

    :param portfolio_id: id of portfolio
    :param rolling_window: size of the rolling window to look back on
    :param benchmark_id: id of benchmark asset
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio r squared
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.r_squared(performance_report.id, benchmark_id, rolling_window)


@plot_measure_entity(EntityType.PORTFOLIO, [QueryType.CAPTURE_RATIO])
def portfolio_capture_ratio(portfolio_id: str, benchmark_id: str, rolling_window: Union[int, str], *,
                            source: str = None,
                            real_time: bool = False, request_id: Optional[str] = None) -> pd.Series:
    """
    A measure of how well portfolio is doing relative to the benchmark,
    defined as the ratio of the portfolio return to the benchmark return calculated
    on a daily basis and averaged over the stated time period.

    :param portfolio_id: id of portfolio
    :param benchmark_id: id of benchmark asset
    :param rolling_window: size of the rolling window to look back on
    :param source: name of function caller
    :param real_time: whether to retrieve intraday data instead of EOD
    :param request_id: server request id
    :return: portfolio capture ratio
    """
    pm = PortfolioManager(portfolio_id)

    performance_report = pm.get_performance_report()
    return ReportMeasures.capture_ratio(performance_report.id, benchmark_id, rolling_window)
